#include "bitmap.h"


Bitmap::Bitmap(int numDimensions, int *dimSizes, int numPoints, int** points){
	init(numDimensions, dimSizes, numPoints, points);	
}

void Bitmap::init(int numDimensions, int *dimSizes, int numPoints, int** points){
	this->numDimensions = numDimensions;
	this->numPoints = numPoints;
	this->dimSizes = dimSizes;
	this->points = points;
	
	sumDimSizes = new int[numDimensions];
	sumDimSizes[0] = 0;
	for (int i=1; i<numDimensions; i++){
		sumDimSizes[i] = sumDimSizes[i-1] + dimSizes[i-1];
	}
	
	length = sumDimSizes[numDimensions-1] + dimSizes[numDimensions-1];
	
	bslices = new boost::dynamic_bitset<>[length];
	for (int i=0; i<length; i++){
		bslices[i] = dbitset(numPoints, 0ul);
	}
	
	for (int p=0; p<numPoints; p++){
		for (int d=0; d<numDimensions; d++){
			int val = points[p][d];
			for (int k=val; k<dimSizes[d]; k++){
				bslices[sumDimSizes[d]+k][p] = true;
			}
			
		}
		
	}
	
}

Bitmap::Bitmap(char *infile){
	fstream fin;
	fin.open(infile, ios::in);
	fin >> numDimensions; // read dimensionality
	dimSizes = new int[numDimensions];
	for (int d=0; d<numDimensions; d++){ // read dimension sizes
		fin >> dimSizes[d];
		dimSizes[d]++;
//		cout << dimSizes[d] << " ";
	}
//	cout << endl;
	fin >> numPoints; // read number of points
	points = new int*[numPoints];
	for (int p=0; p<numPoints; p++){ // read points
		points[p] = new int[numDimensions];
//		cout << "point " << p << ": ";
		for (int d=0; d<numDimensions; d++){
			fin >> points[p][d];
//			cout << points[p][d] << " ";
		}
//		cout << endl;
	}
	fin.close();
//	cout << "numDimensions=" << numDimensions << " numPoints=" << numPoints << endl;
	init(numDimensions, dimSizes, numPoints, points);
}


dbitset Bitmap::isDominated(int *point){
	dbitset a(numPoints);
	a.set();
	dbitset b(numPoints);
	b.reset();
	for (int d=0; d<numDimensions; d++){
		a &= bslices[sumDimSizes[d]+point[d]];
		if (point[d]!=0){
			b |= bslices[sumDimSizes[d]+point[d]-1];
		}
	}
	return a&b;
}

dbitset Bitmap::isDominated(int *point, dbitset mask){
	dbitset a = mask;
	dbitset b(numPoints);
	b.reset();
	for (int d=0; d<numDimensions; d++){
		a &= bslices[sumDimSizes[d]+point[d]];
		if (point[d]!=0){
			b |= bslices[sumDimSizes[d]+point[d]-1];
		}
	}
	return a&b;
}


dbitset Bitmap::dominates(int *point){
	dbitset a(numPoints);
	a.set();
	dbitset b(numPoints);
	b.reset();
	for (int d=0; d<numDimensions; d++){
		a &= bslices[sumDimSizes[d]+point[d]];
		if (point[d]!=0){
			b |= bslices[sumDimSizes[d]+point[d]-1];
		}
	}
	return ~(a|b);
}


dbitset Bitmap::getSkyline(){
	#ifdef PRINT_RESULTS
	cout << endl << "Skyline bitmap method" << endl;
	#endif

	dbitset mask = dbitset(numPoints);
	mask.set();
	return getSkyline(mask);
}

dbitset Bitmap::getSkyline(dbitset mask){
	dbitset result = mask;

//cout << result << endl;

	for (int p=0; p<numPoints; p++){
		if (!result[p]){
			continue;
		}
		dbitset isdom = isDominated(points[p], mask);
		if (isdom.any()){ // not in the skyline
			result.reset(p);
		}
//cout << result << endl;
	}
	return result;
}

dbitset Bitmap::getSkyband(int k){
	dbitset mask = dbitset(numPoints);
	mask.set();
	return getSkyband(k, mask);
}

dbitset Bitmap::getSkyband(int k, dbitset mask){
	dbitset result = mask;
	for (int p=0; p<numPoints; p++){
		if (!result[p]){
			continue;
		}
		dbitset isdom = isDominated(points[p], mask);
		int count = 0;
		for (int i=0; i<numPoints; i++){
			if (isdom[i]){
				count++;
				if (count>=k) break;
			}
		}
		if (count>=k){ // not in the skyband
			result.reset(p);
		}
	}
	return result;
}


dbitset Bitmap::getSkylayer(int l){
	dbitset mask = dbitset(numPoints);
	mask.set();
	return getSkylayer(l, mask);
}

dbitset Bitmap::getSkylayer(int l, dbitset mask){
	dbitset result = dbitset(numPoints);
	result.reset();
	for (int i=0; i<l; i++){
		mask = mask & (~result);
		result.reset();
		for (int p=0; p<numPoints; p++){
			if (!mask[p]){
				continue;
			}
			dbitset isdom = isDominated(points[p], mask);
			if (!isdom.any()){ // it's in the skyline
				result.set(p);
			}
		}
	}
	return result;
}


dbitset Bitmap::getSkylayers(int l){
	dbitset mask = dbitset(numPoints);
	mask.set();
	return getSkylayer(l, mask);
}

dbitset Bitmap::getSkylayers(int l, dbitset mask){
	dbitset result = dbitset(numPoints);
	result.reset();
	for (int i=0; i<l; i++){
		mask = mask & (~result);
		// do not reset result
		for (int p=0; p<numPoints; p++){
			if (!mask[p]){
				continue;
			}
			dbitset isdom = isDominated(points[p], mask);
			if (!isdom.any()){ // it's in the skyline
				result.set(p);
			}
		}
	}
	return result;
}


int Bitmap::getDominanceScore(int *point){
	dbitset doms = dominates(point);
	int score = 0;
	for (int p=0; p<numPoints; p++){
		if (doms.test(p)){
			score++;
		}
	}
	return score;
}


vector<int> Bitmap::getIDsFromBitset(dbitset bitset){
	vector<int> ids;
	for (int p=0; p<numPoints; p++){
		if (bitset.test(p)){
			ids.push_back(p);
		}
	}
	return ids;
}

string Bitmap::printIDsFromBitset(dbitset bitset){
	vector<int> ids;
	for (int p=0; p<numPoints; p++){
		if (bitset.test(p)){
			ids.push_back(p);
		}
	}
	
	ostringstream sout;
	for (uint i=0; i<ids.size(); i++){
		sout << ids[i] << " ";
	}
	return sout.str();
}


string Bitmap::toString(){
	ostringstream sout;
	for (int i=0; i<length; i++){
		sout << "slice " << i << ": " << printSlice(i) << endl;
	}
	return sout.str();
}


string Bitmap::printSlice(int slice){
	ostringstream sout;
	for (int p=0; p<numPoints; p++){
		sout << bslices[slice][p];
	}
	return sout.str();
}


Bitmap::~Bitmap()
{
	delete[] dimSizes;
	delete[] sumDimSizes;
	for (int j = 0; j < numPoints; j++)
		delete[] points[j];
}
