#include "timing.h"
#include "port.h"
#include "naive.h"
#include "BnL.h"
#include "DnC.h"
#include "bitmap.h"
#include "fast_bitmap.h"



const char* progname = 0;

#define warning(FMT, ...) fprintf(stderr, "\n%s warning: " FMT "\n", progname, ##__VA_ARGS__)
#define invalidargs(FMT, ...) do { fprintf(stderr, "\n%s error: " FMT "\n", progname, ##__VA_ARGS__); fprintf(stderr, "\n"); usage(); exit(1); } while (0)
#define fatal(FMT, ...) do { fprintf(stderr, "\n%s error: " FMT "\n", progname, ##__VA_ARGS__); exit(1); } while (0)

void usage()
{
	fprintf(stderr, 
"\n\
Usage: %s (-N|-L|-D|-B|-F) -f INPUT [-w WINDOW] [-l SKYLAYER] -i ITERATIONS\n\n\
Options:\n\
       -N            naive method O(N^2) dominance tests\n\
       -L            block nested loops method\n\
       -D            basic divide and conquer method\n\
       -B            bitmap method\n\
       -F            fast bitmap method\n\
\n\
       -f INPUT      specify input file of points\n\
       -w WINDOW     specify window size (used only in block nested loops and divide and conquer methods)\n\
       -l SKYLAYER   specify skyline layer (used only bitmap-based methods, 1 for basic skyline)\n\
\n\
       -i ITERATIONS specify number of iterations\n\
\n\
Examples:\n\
       %s -N -f iD10T1M7C1000 -i 100\n\
       %s -L -f iD10T1M7C1000 -w 20 -i 100\n\
       %s -D -f iD10T1M7C1000 -w 20 -i 100\n\
       %s -B -f iD10T1M7C1000 -l 1 -i 100\n\
       %s -F -f iD10T1M7C1000 -l 1 -i 100\n\
\n", 
	progname, progname, progname, progname, progname, progname);
}


int main(int argc, char** argv)
{
	/*if (argc > 0)
		progname = basename(argv[0]);
	if ((progname == 0) || (progname[0] == '\0'))*/
	progname = "mainSkylines";

	if (argc < 2)
		invalidargs("too few arguments");

	int c;
	int method = 0;
	char* input_file = 0;
	int window_size = 0;
	int l_skylayer = 0;
	int iterations = 0;
			
	while ((c = getopt(argc, argv, "NLDBFR:f:w:l:i:h?")) != -1)
	{
		// R is dummy argument! WHY DO WE NEED IT?!?!?!
		switch (c)
		{
		case 'N':
		case 'L':
		case 'D':
		case 'B':
		case 'F':
			if ((method != 0) && (method != c))
				invalidargs("method already selected, hint: use only one of -N | -L | -D | -B | -F");
			method = c;

			#ifdef DEBUG
			cout << "Method: " << method << endl;
			#endif
			break;

		case 'R':
			break;

		case 'f':
			input_file = optarg;

			#ifdef DEBUG
			cout << "Input file: " << input_file << endl;
			#endif
			break;

		case 'w':
			window_size = atoi(optarg);

			#ifdef DEBUG
			cout << "Window size: " << window_size << endl;
			#endif
			break;

		case 'l':
			l_skylayer = atoi(optarg);

			#ifdef DEBUG
			cout << "l_skylayer: " << l_skylayer << endl;
			#endif
			break;

		case 'i':
			iterations = atoi(optarg);

			#ifdef DEBUG
			cout << "iterations: " << iterations << endl;
			#endif
			break;

		case 'h':
		case '?':
			usage();
			return 0;

		default:
			usage();
			return 1;
		}
	}
	
	// Check for additional arguments
	/*while (optind < argc)
	{
		warning("ignoring additional argument '%s'", argv[optind]);
		optind++;
	}*/

	// Perform sanity checks on arguments
	if (method == 0)
		invalidargs("no method selected");
	if (input_file == 0)
		invalidargs("no input file of points specified");
	if (iterations <= 0)
		invalidargs("iterations must be > 0");
	if ((method == 'L') && (window_size == 0))
		invalidargs("window size must > 0");
	if (((method == 'B') || (method == 'F')) && (l_skylayer == 0))
		invalidargs("l-skylayer must be >=1");
	if (((method == 'N') || (method == 'D')) && ((window_size != 0) || (l_skylayer != 0)))
		warning("ignoring additional argument 'w', 'l'");
	if ((method == 'L') && (l_skylayer != 0))
		warning("ignoring additional argument 'l'");
	if (((method == 'B') || (method == 'F')) && (window_size != 0))
		warning("ignoring additional argument 'w'");

	Timing* tim = new Timing();
	long duration = 0;
	Naive* naive = 0;
	BlockNestedLoops* bnl = 0;
	DivideAndConquer* dnc = 0;
	Bitmap* bitmap = 0;
	FastBitmap* fast_bitmap = 0;
	vector<int*> bnl_skyline, naive_skyline;
	deque<int*> dnc_skyline;
	dbitset bitmap_skyline, fast_bitmap_skyline;
	int size = 0;

	switch (method)
	{
	case 'N':
		tim->startClock();
		for (int count = 0; count < iterations; count++)
		{
			naive = new Naive(input_file);
			naive_skyline = naive->getSkyline();

			#ifdef PRINT_RESULTS
			for (int p = 0; p < naive_skyline.size(); p++)
			{
				for (int d = 0; d < naive->numDimensions; d++)
					cout << naive_skyline[p][d] << " ";
				cout << endl;
			}
			#endif

			delete naive;
		}
		duration = tim->stopClock();
		cout << "\nskyline size: " << naive_skyline.size() << endl;
		cout << "time elapsed: " << ((float)duration) << endl;
		naive_skyline.clear();
		break;

	case 'L':
		tim->startClock();
		for (int count = 0; count < iterations; count++)
		{
			bnl = new BlockNestedLoops(input_file);
			bnl_skyline = bnl->getSkyline(window_size);

			#ifdef PRINT_RESULTS
			for (int p = 0; p < bnl_skyline.size(); p++)
			{
				for (int d = 1; d < bnl->numDimensions+1; d++)
					cout << bnl_skyline[p][d] << " ";
				cout << endl;
			}
			#endif

			delete bnl;
		}
		duration = tim->stopClock();
		cout << "\nskyline size: " << bnl_skyline.size() << endl;
		cout << "time elapsed: " << ((float)duration) << endl;
		bnl_skyline.clear();
		break;

	case 'D':
		tim->startClock();
		for (int count = 0; count < iterations; count++)
		{
			dnc = new DivideAndConquer(input_file);
			dnc_skyline = dnc->getSkylineBasic(dnc->numDimensions-1);

			#ifdef PRINT_RESULTS
			for (int p = 0; p < dnc_skyline.size(); p++)
			{
				for (int d = 0; d < dnc->numDimensions; d++)
					cout << dnc_skyline[p][d] << " ";
				cout << endl;
			}
			#endif

			delete dnc;
		}
		duration = tim->stopClock();
		cout << "\nskyline size: " << dnc_skyline.size() << endl;
		cout << "time elapsed: " << ((float)duration) << endl;
		dnc_skyline.clear();
		break;

	case 'B':
		tim->startClock();
		for (int count = 0; count < iterations; count++)
		{
			bitmap = new Bitmap(input_file);
			bitmap_skyline = bitmap->getSkyline();

			#ifdef PRINT_RESULTS
			for (unsigned int p = 0; p < bitmap_skyline.size(); p++)
			{
				if (bitmap_skyline[p] == 1)
				{
					for (int d = 0; d < bitmap->numDimensions; d++)
						cout << bitmap->points[p][d] << " ";
					cout << endl;
				}
			}
			#endif

			delete bitmap;
		}
		duration = tim->stopClock();
		for (unsigned int p = 0; p < bitmap_skyline.size(); p++)
		{
			if (bitmap_skyline[p] == 1)
				size++;
		}
		cout << "\nskyline size: " << size << endl;
		cout << "time elapsed: " << ((float)duration) << endl;
		bitmap_skyline.clear();
		break;

	case 'F':
		tim->startClock();
		for (int count = 0; count < iterations; count++)
		{
			fast_bitmap = new FastBitmap(input_file);
			fast_bitmap_skyline = fast_bitmap->getSkyline();

			#ifdef PRINT_RESULTS
			for (unsigned int p = 0; p < fast_bitmap_skyline.size(); p++)
			{
				if (fast_bitmap_skyline[p] == 1)
				{
					for (int d = 0; d < fast_bitmap->numDimensions; d++)
						cout << fast_bitmap->points[p][d] << " ";
					cout << endl;
				}
			}
			#endif
			cout << fast_bitmap_skyline << endl;

			delete fast_bitmap;
		}
		duration = tim->stopClock();
		for (unsigned int p = 0; p < fast_bitmap_skyline.size(); p++)
		{
			if (fast_bitmap_skyline[p] == 1)
				size++;
		}
		cout << "\nskyline size: " << size << endl;
		cout << "time elapsed: " << ((float)duration) << endl;
		fast_bitmap_skyline.clear();
		break;

	default:
		fatal("method `%c' is not supported", method);
		break;
	}

	delete tim;
	
	return 0;	
}
